from collections import defaultdict
from datetime import date

import pytest

from pychores.domain.entity.chore import Chore
from pychores.domain.use_cases.create_chore import (
    CreateChore,
    ICreateChore,
    InvalidChorePayload,
)


class MemRepo(ICreateChore):
    def __init__(self):
        self.chores = defaultdict(list)

    def save_new_chore(self, username: str, chore: Chore):
        self.chores[username].append(chore)


def dp():
    return date(2000, 1, 1)


class TestCreateCore:
    def test_should_prevent_creation_if_period_is_invalid(self):
        payload = {
            "name": "wrong_period_type",
            "description": "A chore withe a wrong period type",
            "period": "str_should_be_int",
        }
        uc = CreateChore(MemRepo(), date_provider=dp)
        with pytest.raises(InvalidChorePayload):
            uc.execute("user1", payload=payload)



"""Bind Chores to user

Revision ID: a6fe0494c1f0
Revises: 944acee085dd
Create Date: 2020-07-29 15:37:16.561484

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "a6fe0494c1f0"
down_revision = "944acee085dd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("chore", schema=None) as batch_op:
        batch_op.add_column(sa.Column("user_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key("chore_user_id_fkey", "user", ["user_id"], ["id"])
        # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("chore_user_id_fkey", "chore", type_="foreignkey")
    op.drop_column("chore", "user_id")
    # ### end Alembic commands ###



"""Bind Chores to user

Revision ID: a6fe0494c1f1
Revises: a6fe0494c1f0
Create Date: 2020-07-29 15:37:16.561484

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "a6fe0494c1f1"
down_revision = "a6fe0494c1f0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("UPDATE chore SET USER_ID = 1")

    with op.batch_alter_table("chore", schema=None) as batch_op:
        batch_op.alter_column("user_id", nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("chore", "user_id", nullable=True)
    # ### end Alembic commands ###



import json
from datetime import date

from flask import url_for

from pychores.model import Chore

from .helpers import get_auth_headers, object_match_dict


class TestChores:
    def test_get_chore_should_return_list_of_chores(self, client, chore_factory):
        chore1 = chore_factory()
        chore2 = chore_factory(user=chore1.user)

        r = client.get(url_for("chores.chores"), headers=get_auth_headers(chore1.user))

        assert r.status_code == 200
        chores = r.json
        assert "chores" in chores
        assert chores["chores"][0]["name"] == chore1.name
        assert chores["chores"][1]["name"] == chore2.name

    def test_chore_should_be_empty_after_insert_in_previous_test(self, db_session):
        chores = db_session.query(Chore).all()
        assert len(chores) == 0

    def test_post_chores_should_add_a_chore(self, client, headers, user, db_session):
        chore_to_post = {
            "name": "a new chore",
            "description": "description of the new chore",
            "period": 9,
        }
        headers.update(get_auth_headers(user))
        r = client.post(
            url_for("chores.chores"), data=json.dumps(chore_to_post), headers=headers
        )

        assert r.status_code == 200
        chore = db_session.query(Chore).filter_by(name=chore_to_post["name"]).one()
        assert object_match_dict(chore, chore_to_post)

    def test_post_chore_should_return_400_on_invalid_payload(
        self, client, headers, user, db_session
    ):
        chore_to_post = {
            "name": "a new chore",
            "description": "description of the new chore",
            "period": "invalid",
        }
        headers.update(get_auth_headers(user))
        r = client.post(
            url_for("chores.chores"), data=json.dumps(chore_to_post), headers=headers
        )

        assert r.status_code == 400

    def test_chore_should_be_empty(self, db_session):
        chores = db_session.query(Chore).all()
        assert len(chores) == 0


class TestChore:
    def test_get_should_return_chore(self, client, chore):
        r = client.get(
            url_for("chores.chore", chore_id=chore.id),
            headers=get_auth_headers(chore.user),
        )
        assert chore.name == r.json["name"]

    def test_put_should_change_chore(self, client, headers, chore):
        chore_to_change = {
            "name": "a new chore",
            "description": "description of the new chore",
            "period": 9,
        }
        headers.update(get_auth_headers(chore.user))
        r = client.put(
            url_for("chores.chore", chore_id=chore.id),
            data=json.dumps(chore_to_change),
            headers=headers,
        )
        assert r.status_code == 200
        assert object_match_dict(chore, chore_to_change)

    def test_delete_chore_should_delete_chore(self, client, chore, db_session):
        r = client.delete(
            url_for("chores.chore", chore_id=chore.id),
            headers=get_auth_headers(chore.user),
        )

        assert r.status_code == 204
        assert db_session.get(Chore, chore.id) is None

    def test_post_should_add_task_with_default(self, client, chore, headers):
        headers.update(get_auth_headers(chore.user))
        r = client.post(
            url_for("chores.chore", chore_id=chore.id),
            headers=headers,
            data=json.dumps({}),
        )

        assert r.status_code == 200
        assert len(chore.tasks) == 1
        assert chore.tasks[0].execution_date == date.today()

    def test_get_wrong_chore_should_abort(self, client, chore_factory):
        chore1 = chore_factory()
        chore2 = chore_factory()

        r = client.get(
            url_for("chores.chore", chore_id=chore1.id),
            headers=get_auth_headers(chore2.user),
        )

        assert r.status_code == 401



from flask_jwt_extended import create_access_token


def object_match_dict(obj, dictionary, whitelist=[]):
    for k, v in dictionary.items():
        if getattr(obj, k) != v and k not in whitelist:
            return False
    return True


def get_auth_headers(user):
    return {"Authorization": f"Bearer {create_access_token(user.username)}"}



from abc import ABC, abstractmethod
from typing import Callable

from pychores.domain.entity.user import User


class IGetUserRepo(ABC):
    @abstractmethod
    def get_user(self, username: str) -> User:
        """Retrieve a user by username"""


class SignUserIn:
    def __init__(
        self, repo: IGetUserRepo, password_checker: Callable[[str, str], bool]
    ):
        self.repo = repo
        self.password_checker = password_checker

    def execute(self, payload: dict) -> User:
        user = self.repo.get_user(payload["username"])
        if not self.password_checker(user.password, payload["password"]):
            raise ValueError("Wrong password or username")
        return user



from abc import ABC, abstractmethod

from pychores.domain.entity.task import Task


class IDeleteTask(ABC):
    @abstractmethod
    def delete_task(self, task: Task):
        """Delete task"""

    @abstractmethod
    def get_task(self, username: str, task_id: int) -> Task:
        """Get task owned by username"""


class DeleteTask:
    def __init__(self, repo: IDeleteTask):
        self.repo = repo

    def execute(self, username: str, task_id: int):
        task = self.repo.get_task(username=username, task_id=task_id)
        self.repo.delete_task(task)



from abc import ABC, abstractmethod

from pychores.domain.entity.chore import Chore


class IEditChore(ABC):
    @abstractmethod
    def get_chore(self, username: str, chore_id: int) -> Chore:
        """Return Chore with chore_id for username if exists"""

    @abstractmethod
    def save_chore(self, chore: Chore) -> Chore:
        """Save the chore"""


class EditChore:
    def __init__(self, repo: IEditChore):
        self.repo = repo

    def execute(self, username: str, chore_id: int, payload: dict) -> Chore:
        old_chore = self.repo.get_chore(username, chore_id)
        new_chore = Chore(
            id=old_chore.id,
            name=payload.get("name", old_chore.name),
            current_date=old_chore.current_date,
            description=payload.get("description", old_chore.description),
            period=payload.get("period", old_chore.period),
            last_execution=old_chore.last_execution,
        )
        self.repo.save_chore(new_chore)
        return new_chore



import pytest
from flask import url_for


@pytest.mark.skip(reason="This should be theorically true, but can't make it work")
def test_session_instance(db_session, chore_factory, task_factory):
    assert db_session == chore_factory._meta.sqlalchemy_session
    assert db_session == task_factory._meta.sqlalchemy_session


def test_version_route(client):
    r = client.get(url_for("api_version"))
    assert r.status_code == 200


def test_static_file(client):
    r = client.get(url_for("catch_all"))
    assert r.status_code == 200



from datetime import date

from flask import url_for

from pychores.model import Task

from .helpers import get_auth_headers


class TestTask:
    def test_get_should_return_task(self, client, task_factory):
        task = task_factory(execution_date=date(2020, 5, 1))
        r = client.get(
            url_for("tasks.task", task_id=task.id),
            headers=get_auth_headers(task.chore.user),
        )
        assert r.json["task"]["execution_date"] == "Fri, 01 May 2020 00:00:00 GMT"
        assert r.json["task"]["chore_id"] == task.chore_id

    def test_delete_should_erase_task(self, client, task, db_session):
        r = client.delete(
            url_for("tasks.task", task_id=task.id),
            headers=get_auth_headers(task.chore.user),
        )
        assert r.status_code == 204
        assert db_session.get(Task, task.id) is None



import json

from flask import url_for

from pychores.adapter.provider.hash import hash_provider
from pychores.model import User

from .helpers import object_match_dict


class TestAuth:
    def test_signup_should_create_user(self, db_session, client, headers):
        user_to_post = {
            "username": "test_a",
            "password": "pass_test_a",
            "email": "test_a@example.com",
        }

        r = client.post(
            url_for("auth.create_user"), data=json.dumps(user_to_post), headers=headers
        )

        assert r.status_code == 200
        user = db_session.query(User).filter_by(username=user_to_post["username"]).one()
        assert object_match_dict(user, user_to_post, ["password"])

    def test_signin_should_return_token(
        self, client, user_factory, headers, db_session
    ):
        password = hash_provider("password")
        user_factory(username="user", password=password)
        signin = {"username": "user", "password": "password"}

        r = client.post(
            url_for("auth.signin"), data=json.dumps(signin), headers=headers
        )

        assert r.status_code == 200
        assert r.json["access_token"]
        assert r.json["username"] == "user"



from pychores.domain.entity.task import Task
from pychores.domain.use_cases.add_task import IAddTask
from pychores.model import Task as DbTask

from .builder import build_chore


class TaskRepo(IAddTask):
    def __init__(self, session):
        self.session = session

    def save_task(self, task: Task) -> Task:
        db_task = DbTask(
            id=task.id,
            chore_id=task.chore.id,
            execution_date=task.execution_date,
        )
        self.session.add(db_task)
        self.session.commit()
        return Task(
            id=db_task.id, execution_date=db_task.execution_date, chore=task.chore
        )

    def get_task(self, username: str, task_id: int) -> Task:
        db_task = self.session.get(DbTask, task_id)
        assert db_task.chore.user.username == username
        return Task(
            execution_date=db_task.execution_date,
            id=db_task.id,
            chore=build_chore(db_task.chore),
        )

    def delete_task(self, task: Task):
        db_task = self.session.get(DbTask, task.id)
        self.session.delete(db_task)
        self.session.commit()



from flask_bcrypt import check_password_hash, generate_password_hash


def hash_provider(raw_string: str) -> str:
    return generate_password_hash(raw_string).decode("utf-8")


def check_provider(encoded_string: str, raw_string: str) -> bool:
    return check_password_hash(encoded_string, raw_string)



from datetime import date

from pychores.domain.entity.chore import Chore
from pychores.model import Chore as DbChore


def build_chore(db_chore: DbChore) -> Chore:
    return Chore(
        id=db_chore.id,
        name=db_chore.name,
        description=db_chore.description,
        period=db_chore.period,
        current_date=date.today(),
        last_execution=max(t.execution_date for t in db_chore.tasks)
        if db_chore.tasks
        else None,
    )



from abc import ABC, abstractmethod
from typing import Callable

from pychores.domain.entity.user import User


class ISaveUserRepo(ABC):
    @abstractmethod
    def save_user(self, user: User):
        """Save the domain part of user"""


class CreateUser:
    def __init__(self, repo: ISaveUserRepo, password_hasher: Callable[[str], str]):
        self.repo = repo
        self.password_hasher = password_hasher

    def execute(self, payload: dict) -> User:
        user = User(
            username=payload["username"],
            email=payload["email"],
            password=self.password_hasher(payload["password"]),
        )
        self.repo.save_user(user)
        return user
